@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1997, 1998, 1999 Kungliga Tekniska Hgskolan
+ * Copyright (c) 1997-1999 Kungliga Tekniska Hgskolan
  * (Royal Institute of Technology, Stockholm, Sweden). 
  * All rights reserved. 
  *
@@ -37,7 +37,7 @@
  */
 
 #include "push_locl.h"
-RCSID("$Id: push.c,v 1.22 1999/02/27 22:45:40 assar Exp $");
+RCSID("$Id: push.c,v 1.31 1999/08/19 17:26:53 assar Exp $");
 
 #ifdef KRB4
 static int use_v4 = -1;
@@ -96,46 +96,55 @@ usage (int ret)
 }
 
 static int
-do_connect (char *host, int port, int nodelay)
+do_connect (const char *hostname, int port, int nodelay)
 {
-    struct hostent *h;
-    struct sockaddr_in addr;
-    char **p;
-    int s = -1;
+    struct hostent *hostent = NULL;
+    char **h;
+    int error;
+    int af;
+    int s;
 
-    h = roken_gethostbyname (host);
-    if (h == NULL)
-	errx (1, "gethostbyname: %s", hstrerror(h_errno));
-    memset (&addr, 0, sizeof(addr));
-    addr.sin_family = AF_INET;
-    addr.sin_port   = port;
-    for (p = h->h_addr_list; *p; ++p) {
-	memcpy(&addr.sin_addr, *p, sizeof(addr.sin_addr));
+#ifdef HAVE_IPV6    
+    if (hostent == NULL)
+	hostent = getipnodebyname (hostname, AF_INET6, 0, &error);
+#endif
+    if (hostent == NULL)
+	hostent = getipnodebyname (hostname, AF_INET, 0, &error);
 
-	s = socket (AF_INET, SOCK_STREAM, 0);
+    if (hostent == NULL)
+	errx(1, "gethostbyname '%s' failed: %s", hostname, hstrerror(error));
+
+    af = hostent->h_addrtype;
+
+    for (h = hostent->h_addr_list; *h != NULL; ++h) {
+	struct sockaddr_storage sa_ss;
+	struct sockaddr *sa = (struct sockaddr *)&sa_ss;
+
+	sa->sa_family = af;
+	socket_set_address_and_port (sa, *h, port);
+
+	s = socket (af, SOCK_STREAM, 0);
 	if (s < 0)
 	    err (1, "socket");
-	if (connect(s, (struct sockaddr *)&addr, sizeof(addr)) < 0) {
-	    warn ("connect(%s)", host);
+	if (connect(s, sa, socket_sockaddr_size(sa)) < 0) {
+	    warn ("connect(%s)", hostname);
 	    close (s);
 	    continue;
 	} else {
 	    break;
 	}
     }
-    if (*p == NULL)
+    freehostent (hostent);
+    if (*h == NULL)
 	return -1;
-    else {
-	if(setsockopt(s, IPPROTO_TCP, TCP_NODELAY,
-		      (void *)&nodelay, sizeof(nodelay)) < 0)
-	    err (1, "setsockopt TCP_NODELAY");
-
-	return s;
-    }
+    if(setsockopt(s, IPPROTO_TCP, TCP_NODELAY,
+		  (void *)&nodelay, sizeof(nodelay)) < 0)
+	err (1, "setsockopt TCP_NODELAY");
+    return s;
 }
 
 typedef enum { INIT = 0, GREET, USER, PASS, STAT, RETR, TOP, 
-	       DELE, QUIT } pop_state;
+	       DELE, XDELE, QUIT} pop_state;
 
 #define PUSH_BUFSIZ 65536
 
@@ -202,10 +211,10 @@ write_state_destroy (struct write_state *w)
 
 static int
 doit(int s,
-     char *host,
-     char *user,
-     char *outfilename,
-     char *header_str,
+     const char *host,
+     const char *user,
+     const char *outfilename,
+     const char *header_str,
      int leavep,
      int verbose,
      int forkp)
@@ -219,6 +228,7 @@ doit(int s,
     pop_state state = INIT;
     unsigned count, bytes;
     unsigned asked_for = 0, retrieved = 0, asked_deleted = 0, deleted = 0;
+    unsigned sent_xdele = 0;
     int out_fd;
     char from_line[128];
     size_t from_line_length;
@@ -260,6 +270,7 @@ doit(int s,
 	FD_SET(s,&readset);
 	if (((state == STAT || state == RETR || state == TOP)
 	     && asked_for < count)
+	    || (state == XDELE && !sent_xdele)
 	    || (state == DELE && asked_deleted < count))
 	    FD_SET(s,&writeset);
 	ret = select (s + 1, &readset, &writeset, NULL, NULL);
@@ -269,22 +280,22 @@ doit(int s,
 	    else
 		err (1, "select");
 	}
-
+	
 	if (FD_ISSET(s, &readset)) {
 	    char *beg, *p;
 	    size_t rem;
 	    int blank_line = 0;
-
+	    
 	    ret = read (s, in_ptr, sizeof(in_buf) - in_len - 1);
 	    if (ret < 0)
 		err (1, "read");
 	    else if (ret == 0)
 		errx (1, "EOF during read");
-
+	    
 	    in_len += ret;
 	    in_ptr += ret;
 	    *in_ptr = '\0';
-
+	    
 	    beg = in_buf;
 	    rem = in_len;
 	    while(rem > 1
@@ -292,9 +303,9 @@ doit(int s,
 		if (state == TOP) {
 		    char *copy = beg;
 
-		    if (strncmp(copy,
-				header_str,
-				min(p - copy + 1, strlen(header_str))) == 0) {
+		    if (strncasecmp(copy,
+				    header_str,
+				    min(p - copy + 1, strlen(header_str))) == 0) {
 			fprintf (stdout, "%.*s\n", (int)(p - copy), copy);
 		    }
 		    if (beg[0] == '.' && beg[1] == '\r' && beg[2] == '\n') {
@@ -326,7 +337,7 @@ doit(int s,
 				    state = QUIT;
 				    net_write (s, "QUIT\r\n", 6);
 				    if (verbose > 1)
-				      net_write (STDERR_FILENO, "QUIT\r\n", 6);
+					net_write (STDERR_FILENO, "QUIT\r\n", 6);
 				} else {
 				    if (forkp) {
 					pid_t pid;
@@ -342,7 +353,7 @@ doit(int s,
 					}
 				    }
 
-				    state = DELE;
+				    state = XDELE;
 				    if (verbose)
 					fprintf (stderr, "deleting... ");
 				}
@@ -369,6 +380,12 @@ doit(int s,
 			    state = TOP;
 			else
 			    state = RETR;
+		    } else if (state == XDELE) {
+			state = QUIT;
+			net_write (s, "QUIT\r\n", 6);
+			if (verbose > 1)
+			    net_write (STDERR_FILENO, "QUIT\r\n", 6);
+			break;
 		    } else if (state == DELE) {
 			if (++deleted == count) {
 			    state = QUIT;
@@ -395,8 +412,14 @@ doit(int s,
 
 		    rem -= p - beg + 2;
 		    beg = p + 2;
-		} else
-		    errx (1, "Bad response: %.*s", (int)(p - beg), beg);
+		} else {
+		    if(state == XDELE) {
+			state = DELE;
+			rem -= p - beg + 2;
+			beg = p + 2;
+		    } else
+			errx (1, "Bad response: %.*s", (int)(p - beg), beg);
+		}
 	    }
 	    if (!do_from)
 		write_state_flush (&write_state);
@@ -412,6 +435,11 @@ doit(int s,
 	    else if ((state == STAT && do_from) || state == TOP)
 		out_len = snprintf (out_buf, sizeof(out_buf),
 				    "TOP %u 0\r\n", ++asked_for);
+	    else if(state == XDELE) {
+		out_len = snprintf(out_buf, sizeof(out_buf),
+				   "XDELE %u %u\r\n", 1, count);
+		sent_xdele++;
+	    }
 	    else if(state == DELE)
 		out_len = snprintf (out_buf, sizeof(out_buf),
 				    "DELE %u\r\n", ++asked_deleted);
@@ -430,17 +458,16 @@ doit(int s,
 
 #ifdef KRB5
 static int
-do_v5 (char *host,
+do_v5 (const char *host,
        int port,
-       char *user,
-       char *filename,
-       char *header_str,
+       const char *user,
+       const char *filename,
+       const char *header_str,
        int leavep,
        int verbose,
        int forkp)
 {
     krb5_error_code ret;
-    krb5_ccache ccache;
     krb5_auth_context auth_context = NULL;
     krb5_principal server;
     int s;
@@ -449,11 +476,6 @@ do_v5 (char *host,
     if (s < 0)
 	return 1;
 
-    ret = krb5_cc_default (context, &ccache);
-    if (ret) {
-	warnx ("krb5_cc_default: %s", krb5_get_err_text (context, ret));
-	return 1;
-    }
     ret = krb5_sname_to_principal (context,
 				   host,
 				   "pop",
@@ -474,12 +496,11 @@ do_v5 (char *host,
 			 0,
 			 NULL,
 			 NULL,
-			 ccache,
+			 NULL,
 			 NULL,
 			 NULL,
 			 NULL);
     krb5_free_principal (context, server);
-    krb5_cc_close (context, ccache);
     if (ret) {
 	warnx ("krb5_sendauth: %s",
 	       krb5_get_err_text (context, ret));
@@ -491,11 +512,11 @@ do_v5 (char *host,
 
 #ifdef KRB4
 static int
-do_v4 (char *host,
+do_v4 (const char *host,
        int port,
-       char *user,
-       char *filename,
-       char *header_str,
+       const char *user,
+       const char *filename,
+       const char *header_str,
        int leavep,
        int verbose,
        int forkp)
@@ -514,7 +535,7 @@ do_v4 (char *host,
 		       s,
 		       &ticket, 
 		       "pop",
-		       host,
+		       (char *)host,
 		       krb_realmofhost(host),
 		       getpid(),
 		       &msg_data,
@@ -531,36 +552,75 @@ do_v4 (char *host,
 }
 #endif /* KRB4 */
 
+#ifdef HESIOD
+
+#ifdef HESIOD_INTERFACES
+
 static char *
-get_pobox (char **user)
+hesiod_get_pobox (const char **user)
+{
+    void *context;
+    struct hesiod_postoffice *hpo;
+    char *ret = NULL;
+
+    if(hesiod_init (&context) != 0)
+	err (1, "hesiod_init");
+
+    hpo = hesiod_getmailhost (context, *user);
+    if (hpo == NULL) {
+	warn ("hesiod_getmailhost %s", *user);
+    } else {
+	if (strcasecmp(hpo->hesiod_po_type, "pop") != 0)
+	    errx (1, "Unsupported po type %s", hpo->hesiod_po_type);
+
+	ret = strdup(hpo->hesiod_po_host);
+	if(ret == NULL)
+	    errx (1, "strdup: out of memory");
+	*user = strdup(hpo->hesiod_po_name);
+	if (*user == NULL)
+	    errx (1, "strdup: out of memory");
+	hesiod_free_postoffice (context, hpo);
+    }
+    hesiod_end (context);
+    return ret;
+}
+
+#else /* !HESIOD_INTERFACES */
+
+static char *
+hesiod_get_pobox (const char **user)
+{
+    char *ret = NULL;
+    struct hes_postoffice *hpo;
+
+    hpo = hes_getmailhost (*user);
+    if (hpo == NULL) {
+	warn ("hes_getmailhost %s", *user);
+    } else {
+	if (strcasecmp(hpo->po_type, "pop") != 0)
+	    errx (1, "Unsupported po type %s", hpo->po_type);
+
+	ret = strdup(hpo->po_host);
+	if(ret == NULL)
+	    errx (1, "strdup: out of memory");
+	*user = strdup(hpo->po_name);
+	if (*user == NULL)
+	    errx (1, "strdup: out of memory");
+    }
+    return ret;
+}
+
+#endif /* HESIOD_INTERFACES */
+
+#endif /* HESIOD */
+
+static char *
+get_pobox (const char **user)
 {
     char *ret = NULL;
 
 #ifdef HESIOD
-    {
-	void *context;
-	struct hesiod_postoffice *hpo;
-
-	if(hesiod_init (&context) != 0)
-	    err (1, "hesiod_init");
-
-	hpo = hesiod_getmailhost (context, *user);
-	if (hpo == NULL) {
-	    warn ("hesiod_getmailhost %s", *user);
-	} else {
-	    if (strcasecmp(hpo->hesiod_po_type, "pop") != 0)
-		errx (1, "Unsupported po type %s", hpo->hesiod_po_type);
-
-	    ret = strdup(hpo->hesiod_po_host);
-	    if(ret == NULL)
-		errx (1, "strdup: out of memory");
-	    *user = strdup(hpo->hesiod_po_name);
-	    if (*user == NULL)
-		errx (1, "strdup: out of memory");
-	    hesiod_free_postoffice (context, hpo);
-	}
-	hesiod_end (context);
-    }
+    ret = hesiod_get_pobox (user);
 #endif
 
     if (ret == NULL)
@@ -571,49 +631,59 @@ get_pobox (char **user)
 }
 
 static void
-parse_pobox (char *a0, char **host, char **user)
+parse_pobox (char *a0, const char **host, const char **user)
 {
-    char *h, *u;
+    const char *h, *u;
+    char *p;
     int po = 0;
 
     if (a0 == NULL) {
-	struct passwd *pwd;
 
-	pwd = getpwuid (getuid ());
-	if (pwd == NULL)
-	    errx (1, "Who are you?");
-	*user = strdup (pwd->pw_name);
-	if (*user == NULL)
-	    errx (1, "strdup: out of memory");
+	*user = getenv ("USERNAME");
+	if (*user == NULL) {
+	    struct passwd *pwd = getpwuid (getuid ());
+
+	    if (pwd == NULL)
+		errx (1, "Who are you?");
+	    *user = strdup (pwd->pw_name);
+	    if (*user == NULL)
+		errx (1, "strdup: out of memory");
+	}
 	*host = get_pobox (user);
 	return;
     }
 
+    /* if the specification starts with po:, remember this information */
     if(strncmp(a0, "po:", 3) == 0) {
 	a0 += 3;
 	po++;
     }
-    h = strchr(a0, '@');
-    if(h)
-	*h++ = '\0';
-    else
+    /* if there is an `@', the hostname is after it, otherwise at the
+       beginning of the string */
+    p = strchr(a0, '@');
+    if(p != NULL) {
+	*p++ = '\0';
+	h = p;
+    } else {
 	h = a0;
-    u = strchr(a0, ':');
-    if(u)
-	*u++ = '\0';
-    else
+    }
+    /* if there is a `:', the username comes before it, otherwise at
+       the beginning of the string */
+    p = strchr(a0, ':');
+    if(p != NULL) {
+	*p++ = '\0';
+	u = p;
+    } else {
 	u = a0;
+    }
     if(h == u) {
 	/* some inconsistent compatibility with various mailers */
 	if(po) {
 	    h = get_pobox (&u);
 	} else {
-	    struct passwd *pwd = getpwuid(getuid());
-	    if (pwd == NULL)
-		errx (1, "Who are you?");
-	    u = strdup(pwd->pw_name);
+	    u = get_default_username ();
 	    if (u == NULL)
-		errx (1, "strdup: out of memory");
+		errx (1, "Who are you?");
 	}
     }
     *host = h;
@@ -626,7 +696,8 @@ main(int argc, char **argv)
     int port = 0;
     int optind = 0;
     int ret = 1;
-    char *host, *user, *filename = NULL, *pobox = NULL;
+    const char *host, *user, *filename = NULL;
+    char *pobox = NULL;
 
     set_progname (argv[0]);
 
@@ -701,7 +772,7 @@ main(int argc, char **argv)
 #ifdef KRB5
 	port = krb5_getportbyname (context, "kpop", "tcp", 1109);
 #elif defined(KRB4)
-	port = k_getportbyname ("kpop", "tcp", 1109);
+    port = k_getportbyname ("kpop", "tcp", 1109);
 #else
 #error must define KRB4 or KRB5
 #endif
